#!/usr/bin/env python3
# Generate backups for various bootstack components.
# This script will create backups in ~/backups of:
# *) Juju configuration
# *) Juju controllers
# *) Swift configuration and rings
# *) MySQL (various OpenStack services)
# *) PostgreSQL (Landscape)
# *) PostgreSQL (MaaS db)

import asyncio
import argparse
import datetime
import glob
import json
import logging
import logging.handlers
import os
import socket
import subprocess
import tarfile
import time
import yaml


class BackupFailed(Exception):
    pass


def juju_model_disable_commands(logger, model_path, disabled_commands):
    if disabled_commands is not None:
        for disabled in disabled_commands:
            command = ['juju', 'disable-command', '-m', model_path, disabled]
            try:
                subprocess.check_call(command)
            except subprocess.CalledProcessError as e:
                logger.debug("Failed to re-disable command {} on juju: "
                             "{}".format(disabled, e))


def get_juju_disabled_commands(logger, model_path):
    try:
        command = ['juju', 'disabled-commands', '-m', model_path]
        output = subprocess.check_output(command)
    except subprocess.CalledProcessError as e:
        logger.debug("Failed to read disabled-commands on juju: "
                     "{}".format(e))
        return(None)

    """ Output like:
        #Disabled commands  Message
        #all
        #destroy-model
        #<blank line>
    Or:
        #No commands are currently disabled.
    """
    disabled_commands = []
    for line in output.splitlines():
        if line == "No commands are currently disabled.":
            return(None)
        elif line == "Disabled commands  Message":
            continue
        elif line == "":
            continue
        else:
            disabled_commands.append(line)

    return(disabled_commands)


def enable_juju_commands(logger, model_path):
    try:
        command = ['juju', 'enable-command', '-m', model_path, 'all']
        output = subprocess.check_output(command)
        logger.debug("Juju commands enabled, output {}".format(output))
    except subprocess.CalledProcessError as e:
        logger.error("Failed to enable-command all on juju: "
                     "{}".format(e))

    return(None)


def get_controllers_list(logger):
    """Return a list of all Juju controllers known to the client

    Needed because libjuju doesn't yet implement the Juju.get_controllers() method
    """
    try:
        command = ['juju', 'controllers', '--format', 'json']
        raw_output = subprocess.check_output(command).decode('UTF-8')
    except subprocess.CalledProcessError as e:
        logger.debug("Failed to read juju controller list: "
                     "{}".format(e))
        return(None)
    controllers = json.loads(raw_output)
    return controllers['controllers'].keys()


async def juju_backup(logger, path, date):
    '''Backup the juju controller database

    Obtain a list of all Juju controllers, loop through each and make a backup of the controller model
    '''
    model = 'controller'
    controller_list = get_controllers_list(logger)
    for current_controller in controller_list:

        filename = "{}/juju-backup-{}-{}.tar.gz".format(path, current_controller, date)
        # as of 14/02/2017, python-libjuju's create_backup is a stub
        # shell out using asyncio subprocess for now
        command = "juju create-backup -m {}:{} " \
                  "--filename {}".format(current_controller, model, filename)
        logger.debug("Backing up Juju controller {}/{}: {}"
                     "".format(current_controller, model, command))

        backup_proc = await asyncio.create_subprocess_shell(command)
        await backup_proc.wait()

        # Remove the backups from the juju controller itself
        remove_command = "for backup in " \
                         "$(juju list-backups -m {0}:{1} | sort | head -n -5); " \
                         "do juju remove-backup $backup -m {0}:{1}; done" \
                         .format(current_controller, model)
        logger.debug("Removing old Juju backups from the controller: {}"
                     "".format(remove_command))

        remove_backup_proc = await asyncio.create_subprocess_shell(remove_command)
        await remove_backup_proc.wait()
    return True


async def juju_client_backup(logger, path, date):
    '''Backup the juju client environment (~/.local/share/juju)
    '''

    filename = "{}/juju-client-backup-{}.tar.gz".format(path, date)

    logger.debug("Backing up juju client settings to {}".format(filename))
    if os.environ.get('JUJU_DATA'):
        clientpath = os.environ['JUJU_DATA']
    else:
        clientpath = os.path.expanduser("~/.local/share/juju")
    with tarfile.open(filename, "w:gz") as tar:
        tar.add(clientpath)

    logger.debug("Juju client settings backed up")


async def aws_client_backup(logger, path, date):

    filename = "{}/aws-client-backup-{}.tar.gz".format(path, date)

    logger.debug("Backing up aws client settings to {}".format(filename))
    if os.environ.get('AWS_CONFIG_FILE'):
        clientpath = os.environ['AWS_CONFIG_FILE'].rsplit('/', 1)[0]
    else:
        clientpath = os.path.expanduser("~/.aws")
    with tarfile.open(filename, "w:gz") as tar:
        tar.add(clientpath)

    logger.debug("AWS client settings backed up")


async def k8s_client_backup(logger, path, date):

    filename = "{}/k8s-client-backup-{}.tar.gz".format(path, date)

    logger.debug("Backing up k8s client settings to {}".format(filename))
    if os.environ.get('KUBECONFIG'):
        clientpath = os.environ['KUBECONFIG'].split(':', 1)[0]
    else:
        clientpath = os.path.expanduser("~/.kube")
    with tarfile.open(filename, "w:gz") as tar:
        tar.add(clientpath)

    logger.debug("k8s client settings backed up")


def get_leader_unit(application, model_path):
    """Given a named application, return the name of the unit which is
    currently the leader.

    :returns string
    """
    # Connect to juju, get an instance of the swift-proxy unit
    command = ["juju", "status", '-m', model_path, application, "--format", "json"]
    tree = subprocess.check_output(command).decode('UTF-8')
    app_stat = json.loads(tree)
    app_units = [unit_name for unit_name in
                 app_stat['applications'][application]['units']]
    for unit in app_units:
        if app_stat['applications'][application]['units'][unit].get('leader'):
            return unit
    return None


async def swift_backup(logger, model_path, path, date):
    """Backup the /etc/swift directory

    TODO make it check the result
    """
    unit_name = get_leader_unit('swift-proxy', model_path)

    cmd = ['juju ssh -m {}, {}'.format(model_path, unit_name),
           'sudo tar zcf /tmp/swift-proxy-{}.tar.gz /etc/swift'.format(date)]
    logger.debug("Backing up swift-proxy: {}".format(cmd))
    proc = await asyncio.create_subprocess_shell(' '.join(cmd))
    await proc.wait()

    cmd = 'juju scp -m {} {}:/tmp/swift-proxy-{}.tar.gz {}/'.format(model_path, unit_name, date, path)
    logger.debug("Downloading the swift-proxy backup: {}".format(cmd))
    proc = await asyncio.create_subprocess_shell(cmd)
    await proc.wait()

    cmd = 'juju ssh -m {} {} sudo rm /tmp/swift-proxy-{}.tar.gz'.format(model_path, unit_name, date)
    logger.debug('Cleaning up swift-proxy backup from unit {}: {}'
                 ''.format(unit_name, cmd))
    proc = await asyncio.create_subprocess_shell(cmd)
    await proc.wait()

    logger.debug("swift-proxy backed up")


async def maas_db_backup(logger, path, date):
    """Backup the psql db from this local maas server

    need sudo/root to -U postgres.
    To add sudo perms:
       echo "jujumanage ALL=(postgres) NOPASSWD:/usr/bin/pg_dumpall" \
       > /etc/sudoers.d/99-jujumanage-backups

    Needs to have this permission in sudoers - do not give jujumanage full sudo
    access!
    """
    command = ['sudo', 'crm_attribute', '--node', socket.gethostname(),
               '--type', 'status', '--name', 'pgsql-status', '--query']
    try:
        resource_info = subprocess.check_output(command)
        if 'value=PRI' in resource_info.decode('UTF-8'):
            run_backup = True
        else:
            run_backup = False
    except subprocess.CalledProcessError as e:
        logging.debug("Failed to check Pacemaker resource location with {}".format(e))
        # Checking for the Postgres corosync state is to avoid duplication, we should try to make a backup anyway
        run_backup = True
    if run_backup:
        command = ["sudo", "-u", "postgres", "pg_dumpall"]
        logger.debug("Backing up maasdb: {}".format(command))
        try:
            with open("{}/maasdb-{}".format(path, date), 'wb') as backup_file:
                subprocess.run(command, stdout=backup_file, check=True)
        except subprocess.CalledProcessError as e:
            logger.error("Maas backup failed with {}".format(e))
        try:
            command = ["gzip", "{}/maasdb-{}".format(path, date)]
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            logger.error("Failed to gzip Maas backup with {}".format(e))
        logger.debug("maasdb backup process complete")
        statinfo = os.stat("{}/maasdb-{}.gz".format(path, date))
        if statinfo.st_size < 100:  # arbitrary that something > 100 bytes is a useful backup
            logger.warn("Maas db backup resulted in a tiny file")
            raise BackupFailed("Maas db backup resulted in a tiny file")
    else:
        # we have Pacemaker, but the VIP is not on this host, so skip backup
        logger.debug("maasdb backup process did not complete as this is not the primary host")


async def mysql_backup(logger, model_path, path, date):
    """
    Backup the leader mysql unit to a local file.
    """
    unit_name = get_leader_unit('mysql', model_path)
    command = ["juju", "run-action", "-m", model_path, "--wait",
               "{}".format(unit_name), "backup",
               "compress=True"]
    logger.debug("Backing up mysql with {}...".format(" ".join(command)))
    backup_proc = \
        asyncio.create_subprocess_shell(" ".join(command),
                                        stdout=asyncio.subprocess.PIPE)
    proc = await backup_proc
    await proc.wait()
    logger.debug("mysql backed up on unit {}".format(unit_name))
    remote_command = ("'sudo tar -zcf /opt/backups/mysqlbackup.tar.gz "
                      "--remove-files /opt/backups/mysql'")
    tar_command = "juju ssh -m {} {} {}".format(model_path, unit_name, remote_command)
    tar_proc = asyncio.create_subprocess_shell(tar_command)
    proc = await tar_proc
    await proc.wait()
    command = ["juju", "scp", "-m", model_path,
               "{}:/opt/backups/mysqlbackup.tar.gz".format(unit_name),
               "{}/mysql-{}.tar.gz".format(path, date)]
    subprocess.run(command, check=True)


async def etcd_backup(logger, model_path, path, date):
    """
    Backup the leader etcd unit to a local file.
    """
    backup_dir = "/home/ubuntu/etcd-backups"
    save_file = "{}/etcd-{}.tar.gz".format(path, date)
    unit_name = get_leader_unit('etcd', model_path)
    command = ["juju", "run-action", "-m", model_path, "--wait",
               "{}".format(unit_name), "snapshot",
               "target={}".format(backup_dir)]
    logger.debug("Backing up etcd with {}...".format(" ".join(command)))
    backup_proc = asyncio.create_subprocess_shell(
            " ".join(command),
            stdout=asyncio.subprocess.PIPE)
    proc = await backup_proc
    await proc.wait()
    output = yaml.load(await proc.stdout.read())
    snapshot_filepath = None
    for unit in output.keys():
        if output[unit]['status'] == 'completed':
            snapshot_filepath = output[unit]['results']['snapshot']['path']
            logger.debug("etcd backed up on unit {}".format(unit_name))
    if snapshot_filepath is not None:
        command = ["juju", "scp", "-m", model_path,
                   "{}:{}".format(unit_name, snapshot_filepath), save_file]

        try:
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            logger.debug("failed to pull backup from unit {}:{} to {}. "
                         "Command: {} ERROR: {}".format(unit_name,
                                                        snapshot_filepath,
                                                        save_file,
                                                        command, e))

        command = ["juju", "run", "-m", model_path, "--unit", unit_name,
                   "rm {}".format(snapshot_filepath)]
        subprocess.run(command, check=True)
    else:
        logger.debug("etcd failed to backup on unit {}: "
                     "OUTPUT: {}".format(unit_name, output))


def send_backups_offsite(logger, host, path, date, cloud_name):
    """
    Find all the files that match the datestamp, and send them to the offsite
    host.
    """
    remote_path = "bootstack@{}:/srv/bootstack-backups".format(host)
    os.chdir(path)
    for filename in glob.glob("*{}*".format(date)):
        command = ['rsync', '-av', filename,
                   "{}/{}-{}".format(remote_path, cloud_name, filename)]
        try:
            logger.debug("Uploading {}: {}".format(filename, ' '.join(command)))
            subprocess.check_output(command).decode('UTF-8')
        except Exception as err:  # pylint: disable=broad-except
            logger.warning("Upload of {} failed with {}".format(filename, err))


def cleanup(logger, path):
    """
    Clean out any old files in the backup path.
    """
    current_time = time.time()
    for backup_file in os.listdir(path):
        backup_file_stat = os.stat(os.path.join(path, backup_file))
        one_week_s = 7 * 24 * 60 * 60
        if backup_file_stat.st_mtime < current_time - one_week_s:
            os.unlink(os.path.join(path, backup_file))
            logger.debug('{} removed'.format(backup_file))


async def get_model_path(args):
    """
    Return a string for use with the -m switch on juju commands

    uses args.controller and args.model to assemble the result, otherwise
    checks the current model and controller and returns that
    :param args: argparse object
    :return: string of the format 'controller:model'
    """
    if args.controller:
        controller_name = "{}:".format(args.controller)
    else:
        controller_name = None
    if args.model:
        model_name = args.model
    else:
        # get current model name
        #
        command = ['juju', 'models']
        if controller_name:
            command.extend('-c', controller_name)
        command.extend(['--format', 'json'])
        try:
            output = subprocess.check_output(command)
        except subprocess.CalledProcessError as e:
            logging.ERROR("failed to list Juju models with {}".format(e))
        models = json.loads(output.decode('UTF-8'))
        model_name = models['current-model']
    if controller_name:
        return '{}{}'.format(controller_name, model_name)
    else:
        return model_name


def main():  # pylint: disable=missing-docstring
    """Main execution

    TODO add relations to automate things which components to backup from which models, and on the leader unit
    make a crontab to execute the right backups.  Meantime just create the crontab manually.
    """
    parser = argparse.ArgumentParser(description='Backup key services '
                                                 'for BootStack.')
    parser.add_argument('--path', dest='path',
                        default='/home/jujumanage/bootstack-backups',
                        help='Path to store backups')
    parser.add_argument('--swift', dest='swift', action='store_true',
                        help='Backup up swift-proxy config')
    parser.add_argument('--mysql', dest='mysql', action='store_true',
                        help='Backup mysql dbs')
    parser.add_argument('--offsite', dest='offsite', action='store_true',
                        help='Send backups to remote host')
    parser.add_argument('--offsite-host', dest='offsite_host',
                        default='bootstack-infra.canonical.com',
                        help='Remote host to use for backup')
    parser.add_argument('--cloud-name', dest='cloud_name',
                        help='Name for this cloud')
    parser.add_argument('--juju', dest='juju',
                        action='store_true',
                        help='Backup Juju controller model and client')
    parser.add_argument('--k8s', dest='k8s',
                        action='store_true',
                        help='Backup Kubernetes etcd and client')
    parser.add_argument('--aws', dest='aws',
                        action='store_true',
                        help='Backup aws client')
    parser.add_argument('--maas', dest='maas',
                        action='store_true',
                        help='Backup maas db')
    parser.add_argument('--controller', dest='controller', default=None,
                        help='Juju controller to use (not relevant for juju backups')
    parser.add_argument('--model', dest='model', default=None,
                        help='Juju model to use (not relevant for juju backups')
    args = parser.parse_args()
    logger = logging
    logger.basicConfig(format='%(levelname)s: %(module)s.%(funcName)s: %(message)s',
                       level=logging.DEBUG)
    # make sure backup directory exists
    args.path = os.path.expanduser(args.path)
    os.makedirs(args.path, int(0o770), True)

    datestamp = "{:%b%d-%Y}".format(datetime.date.today())
    loop = asyncio.get_event_loop()
    model_path = loop.run_until_complete(get_model_path(args))
    logger.debug("Backing up Juju disabled-commands...")
    previously_disabled_commands = get_juju_disabled_commands(logger, model_path)
    enable_juju_commands(logger, model_path)
    task_list = []
    if args.juju:
        task_list.append(juju_backup(logger, args.path, datestamp))
        task_list.append(juju_client_backup(logger, args.path, datestamp))
    if args.swift:
        task_list.append(swift_backup(logger, model_path, args.path, datestamp))
    if args.mysql:
        task_list.append(mysql_backup(logger, model_path, args.path, datestamp))
    if args.maas:
        task_list.append(maas_db_backup(logger, args.path, datestamp))
    if args.aws:
        task_list.append(aws_client_backup(logger, args.path, datestamp))
    if args.k8s:
        task_list.append(etcd_backup(logger, model_path, args.path, datestamp))
        task_list.append(k8s_client_backup(logger, args.path, datestamp))
    loop.run_until_complete(asyncio.gather(*task_list))
    loop.close()
    logger.debug("Restoring up Juju disabled-commands...")
    juju_model_disable_commands(logger, model_path, previously_disabled_commands)
    if args.offsite:
        if args.cloud_name is None:
            raise ValueError('Must set --cloud-name if offsite enabled')
        send_backups_offsite(logger, args.offsite_host, args.path,
                             datestamp, args.cloud_name)
    cleanup(logger, args.path)


if __name__ == '__main__':
    main()
