import glob
import os
import shutil
import socket

from charmhelpers.core import hookenv, host
from charmhelpers.core.templating import render
from charmhelpers.contrib.charmsupport import nrpe


class InfranodeHelper():
    def __init__(self):
        self.charm_config = hookenv.config()

    def action_function(self):
        ''' An example function for calling from an action '''
        return

    def add_nrpe_pacemaker(self, nrpe_setup, hostname):
        """Create NRPE checks for Corosync/Pacemaker

        :param nrpe_setup: nrpe.NRPE object
        :param hostname: hostname defined in nrpe
        :return: None
        """
        scripts_src = os.path.join(os.environ["CHARM_DIR"], "files",
                                   "nrpe")
        scripts_dst = "/usr/local/lib/nagios/plugins"
        self._deploy_files(scripts_src, scripts_dst)
        # corosync/crm checks
        nrpe_setup.add_check(
            'corosync_rings',
            'Check Corosync rings {}'.format(hostname),
            'check_corosync_rings')
        nrpe_setup.add_check(
            'crm_status',
            'Check crm status {}'.format(hostname),
            'check_crm')

        # process checks
        nrpe_setup.add_check(
            'corosync_proc',
            'Check Corosync process {}'.format(hostname),
            'check_procs -c 1:1 -C corosync'
        )
        nrpe_setup.add_check(
            'pacemakerd_proc',
            'Check Pacemakerd process {}'.format(hostname),
            'check_procs -c 1:1 -C pacemakerd'
        )
        nrpe_setup.write()

    def add_nrpe_disabled_juju(self, nrpe_setup):
        """Create NRPE checks for Juju commands

        We try to run Juju models with all commands disabled for safety.  This adds a NRPE check to ensure that is
        the case, will alert if commands are enabled.

        :param nrpe_setup: nrpe.NRPE object
        :return: None
        """
        nrpe_setup.add_check(
            'juju_disabled_commands',
            'Check that Juju commands are disabled',
            '/usr/local/lib/nagios/plugins/check_juju_disabled_commands -f /home/jujumanage/juju_commands_state.txt'
        )
        nrpe_setup.write()
        command = "/snap/bin/juju disabled-commands > ${HOME?}/juju_commands_state.txt"
        cronjob_cmd = (" * * * * * jujumanage timeout -k 10s -s SIGINT 60 "
                       "{command} 2>&1 | logger -p local0.notice\n")
        cronjob = cronjob_cmd.format(command=command)
        host.write_file('/etc/cron.d/juju_disabled_commands', cronjob)

    def create_nrpe_checks(self, nrpe_setup, hostname, current_unit):
        hookenv.log('Running create_nrpe_checks')
        config = hookenv.config()
        services = ['maas-proxy', 'maas-rackd', 'bind9']
        if config.get('nagios_check_dhcpd'):
            services.append('maas-dhcpd')
        nrpe.add_init_service_checks(nrpe_setup, services, current_unit)
        nrpe_setup.write()

        if config.get('nagios_alert_if_juju_enabled'):
            self.add_nrpe_disabled_juju(nrpe_setup)
        if config.get('nagios_check_pacemaker'):
            self.add_nrpe_pacemaker(nrpe_setup, hostname)
        hookenv.status_set('active', 'NRPE configured')

    def deploy_scripts(self):
        """Deploy various scripts and utilities

        Deploys scripts and sudoers pre-requisites for running backups, etc.
        """
        hookenv.log('Running deploy_scripts')
        scripts_dst = '/usr/local/bin'
        scripts_src = os.path.join(os.environ["CHARM_DIR"], "files", "scripts")
        self._deploy_files(scripts_src, scripts_dst)

    def deploy_sudoers(self):
        """Deploy sudoers files to target

        :return: None
        """
        # files
        hookenv.log('Running deploy_sudoers')
        sudoers_src = os.path.join(os.environ["CHARM_DIR"], "files", "sudoers")
        sudoers_dst = "/etc/sudoers.d"
        self._deploy_files(sudoers_src, sudoers_dst)
        # templates
        templates_dir = os.path.join(os.environ["CHARM_DIR"], "templates", "sudoers")
        dest = "/etc/sudoers.d/jujumanage"
        context = {'hostname': socket.gethostname()}
        try:
            render('jujumanage.j2', dest, context, owner='root', group='root',
                   perms=0o644, templates_dir=templates_dir)
        except Exception as e:
            hookenv.log('Deploying template for {} failed with {}'.format(dest, e))
        hookenv.log('Deployed {}'.format(dest))

    def deploy_maasbackup_cron(self):
        """Deploy crontab template for maas backups

        :return: True (i.e. it ran) TODO make that meaningful
        """
        hookenv.log('Running deploy_maasbackup_cron')
        if self.charm_config.get('backup_maas'):
            templates_dir = os.path.join(os.environ["CHARM_DIR"], "templates", "cron")
            cron_dst = "/etc/cron.d/maasbackup"
            context = {'nagios_context': self.charm_config.get('nagios_context'),
                       'offsite': self.charm_config.get('backup_store_offsite')}
            render('maasbackup.j2', cron_dst, context, owner='root', group='root',
                   perms=0o644, templates_dir=templates_dir)
            hookenv.log('Deployed {}'.format(cron_dst))
        return True

    def _deploy_files(self, source, dest):
        """Deploy items from source location to dest

        :param source: string, where to copy files from
        :param dest: string, where to copy files to
        :return: True (i.e. it ran) TODO make that meaningful
        """
        try:
            if not os.path.exists(dest):
                os.makedirs(dest)
            for fname in glob.glob(os.path.join(source, "*")):
                if os.path.isfile(fname):
                    shutil.copy2(fname, os.path.join(dest, os.path.basename(fname)))
        except Exception as e:
            hookenv.log('Deploying file {} failed with {}'.format(dest, e))
        hookenv.log('Deployed {}'.format(dest))
        return True
